# Python Scale Codec
#
# Copyright 2018-2020 openAware BV (NL).
# This file is part of Polkascan.
#
# Polkascan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polkascan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polkascan. If not, see <http://www.gnu.org/licenses/>.
#
#  ss58.py

""" SS58 is a simple address format designed for Substrate based chains.
    Encoding/decoding according to specification on https://wiki.parity.io/External-Address-Format-(SS58)

"""
import base58
from hashlib import blake2b

from scalecodec import ScaleBytes
from scalecodec.types import U8, U16, U32, U64


def ss58_decode(address, valid_address_type=None):
    checksum_prefix = b'SS58PRE'

    ss58_format = base58.b58decode(address)

    if valid_address_type and ss58_format[0] != valid_address_type:
        raise ValueError("Invalid Address type")

    # Determine checksum length according to length of address string
    if len(ss58_format) in [3, 4, 6, 10]:
        checksum_length = 1
    elif len(ss58_format) in [5, 7, 11, 35]:
        checksum_length = 2
    elif len(ss58_format) in [8, 12]:
        checksum_length = 3
    elif len(ss58_format) in [9, 13]:
        checksum_length = 4
    elif len(ss58_format) in [14]:
        checksum_length = 5
    elif len(ss58_format) in [15]:
        checksum_length = 6
    elif len(ss58_format) in [16]:
        checksum_length = 7
    elif len(ss58_format) in [17]:
        checksum_length = 8
    else:
        raise ValueError("Invalid address length")

    checksum = blake2b(checksum_prefix + ss58_format[0:-checksum_length]).digest()

    if checksum[0:checksum_length] != ss58_format[-checksum_length:]:
        raise ValueError("Invalid checksum")

    return ss58_format[1:len(ss58_format)-checksum_length].hex()


def ss58_encode(address, address_type=42):
    checksum_prefix = b'SS58PRE'

    if type(address) is bytes or type(address) is bytearray:
        address_bytes = address
    else:
        address_bytes = bytes.fromhex(address)

    if len(address_bytes) == 32:
        # Checksum size is 2 bytes for public key
        checksum_length = 2
    elif len(address_bytes) in [1, 2, 4, 8]:
        # Checksum size is 1 byte for account index
        checksum_length = 1
    else:
        raise ValueError("Invalid length for address")

    address_format = bytes([address_type]) + address_bytes
    checksum = blake2b(checksum_prefix + address_format).digest()

    return base58.b58encode(address_format + checksum[:checksum_length]).decode()


def ss58_encode_account_index(account_index, address_type=42):

    if 0 <= account_index <= 2**8 - 1:
        account_idx_encoder = U8()
    elif 2**8 <= account_index <= 2**16 - 1:
        account_idx_encoder = U16()
    elif 2**16 <= account_index <= 2**32 - 1:
        account_idx_encoder = U32()
    elif 2**32 <= account_index <= 2**64 - 1:
        account_idx_encoder = U64()
    else:
        raise ValueError("Value too large for an account index")

    return ss58_encode(account_idx_encoder.encode(account_index).data, address_type)


def ss58_decode_account_index(address, valid_address_type=None):

    account_index_bytes = ss58_decode(address, valid_address_type)

    if len(account_index_bytes) == 2:
        return U8(ScaleBytes('0x{}'.format(account_index_bytes))).decode()
    if len(account_index_bytes) == 4:
        return U16(ScaleBytes('0x{}'.format(account_index_bytes))).decode()
    if len(account_index_bytes) == 8:
        return U32(ScaleBytes('0x{}'.format(account_index_bytes))).decode()
    if len(account_index_bytes) == 16:
        return U64(ScaleBytes('0x{}'.format(account_index_bytes))).decode()
    else:
        raise ValueError("Invalid account index length")




#  Scale Codec
#  Copyright (C) 2019  openAware B.V.
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from hashlib import blake2b
from collections import OrderedDict

from scalecodec.base import ScaleDecoder, ScaleBytes
from scalecodec.metadata import MetadataDecoder
from scalecodec.types import Vec, CompactU32, Enum, Bytes, Struct
from scalecodec.utils.ss58 import ss58_decode, ss58_decode_account_index


class ExtrinsicsDecoder(ScaleDecoder):
    type_mapping = (
        ('extrinsic_length', 'Compact<u32>'),
        ('version_info', 'u8'),
        ('address', 'Address'),
        ('signature', 'Signature'),
        ('nonce', 'Compact<u32>'),
        ('era', 'Era'),
        ('call_index', '(u8,u8)'),
    )

    def __init__(self, data=None, sub_type=None, metadata: MetadataDecoder = None, address_type=42):

        assert (type(metadata) == MetadataDecoder)

        self.metadata = metadata
        self.address_type = address_type
        self.extrinsic_length = None
        self.extrinsic_hash = None
        self.version_info = None
        self.contains_transaction = False
        self.address = None
        self.signature_version = None
        self.signature = None
        self.nonce = None
        self.era = None
        self.tip = None
        self.call_index = None
        self.call_module = None
        self.call = None
        self.call_args = None
        self.params_raw = None
        self.params = []
        super().__init__(data, sub_type)

    def generate_hash(self):
        if self.contains_transaction:
            return blake2b(self.data.data, digest_size=32).digest().hex()

    def process(self):
        # TODO for all attributes

        attribute_types = OrderedDict(self.type_mapping)

        self.extrinsic_length = self.process_type('Compact<u32>').value

        if self.extrinsic_length != self.data.get_remaining_length():
            # Fallback for legacy version
            self.extrinsic_length = None
            self.data.reset()

        self.version_info = self.get_next_bytes(1).hex()

        self.contains_transaction = int(self.version_info, 16) >= 80

        if self.version_info == '01' or self.version_info == '81':

            if self.contains_transaction:
                self.address = self.process_type('Address')

                self.signature = self.process_type('Signature').value

                self.nonce = self.process_type(attribute_types['nonce'])

                self.era = self.process_type('Era')

                self.extrinsic_hash = self.generate_hash()

            self.call_index = self.get_next_bytes(2).hex()

        elif self.version_info == '02' or self.version_info == '82':

            if self.contains_transaction:
                self.address = self.process_type('Address')

                self.signature = self.process_type('Signature').value

                self.era = self.process_type('Era')

                self.nonce = self.process_type('Compact<Index>')

                self.tip = self.process_type('Compact<Balance>')

                self.extrinsic_hash = self.generate_hash()

            self.call_index = self.get_next_bytes(2).hex()

        elif self.version_info == '03' or self.version_info == '83':

            if self.contains_transaction:
                self.address = self.process_type('Address')

                self.signature = self.process_type('Signature').value

                self.era = self.process_type('Era')

                self.nonce = self.process_type('Compact<Index>')

                self.tip = self.process_type('Compact<Balance>')

                self.extrinsic_hash = self.generate_hash()

            self.call_index = self.get_next_bytes(2).hex()

        elif self.version_info == '04' or self.version_info == '84':

            if self.contains_transaction:
                self.address = self.process_type('Address')

                multi_signature = self.process_type("MultiSignature")

                self.signature_version = multi_signature.index

                self.signature = multi_signature.get_enum_value()

                self.era = self.process_type('Era')

                self.nonce = self.process_type('Compact<Index>')

                self.tip = self.process_type('Compact<Balance>')

                self.extrinsic_hash = self.generate_hash()

            self.call_index = self.get_next_bytes(2).hex()
        else:
            raise NotImplementedError('Extrinsics version "{}" is not implemented'.format(self.version_info))

        if self.call_index:

            self.params_raw = self.data.data[self.data.offset:]

            # Decode params

            self.call = self.metadata.call_index[self.call_index][1]
            self.call_module = self.metadata.call_index[self.call_index][0]

            for arg in self.call.args:

                arg_type_obj = self.process_type(arg.type, metadata=self.metadata)

                self.params.append({
                    'name': arg.name,
                    'type': arg.type,
                    'value': arg_type_obj.serialize(),
                    'valueRaw': arg_type_obj.raw_value
                })

        result = {
            'valueRaw': self.raw_value,
            'extrinsic_length': self.extrinsic_length,
            'version_info': self.version_info,
        }

        if self.contains_transaction:
            result['account_length'] = self.address.account_length
            result['account_id'] = self.address.account_id
            result['account_index'] = self.address.account_index
            result['account_idx'] = self.address.account_idx
            result['signature_version'] = self.signature_version
            result['signature'] = self.signature.replace('0x', '')
            result['extrinsic_hash'] = self.extrinsic_hash
        if self.call_index:
            result['call_code'] = self.call_index
            result['call_function'] = self.call.get_identifier()
            result['call_module'] = self.call_module.get_identifier()

        if self.nonce:
            result['nonce'] = self.nonce.value

        if self.era:
            result['era'] = self.era.value

        if self.tip:
            result['tip'] = self.tip.value

        result['params'] = self.params

        return result

    def process_encode(self, value):

        if 'call_index' in value:
            self.call_index = value['call_index']

        elif 'call_module' in value and 'call_function' in value:
            # Look up call module from metadata
            for call_index, (call_module, call) in self.metadata.call_index.items():
                if call_module.name.lower() == value['call_module'].lower() and call.name == value['call_function']:
                    self.call_index = call_index
                    self.call_module = call_module
                    self.call = call
                    break

            if not self.call_index:
                raise ValueError('Specified call module and function not found in metadata')

        elif not self.call_module or not self.call:
            raise ValueError('No call module and function specified')

        # Determine version (Fixed to V4 for now)
        if 'account_id' in value:
            self.version_info = '84'
            self.contains_transaction = True
        else:
            self.version_info = '04'
            self.contains_transaction = False

        if self.contains_transaction:
            data = ScaleBytes('0x84')

            self.address = self.get_decoder_class('Address', metadata=self.metadata)
            data += self.address.encode(value['account_id'])

            self.signature_version = self.get_decoder_class('U8', metadata=self.metadata)
            data += self.signature_version.encode(value['signature_version'])

            self.signature = self.get_decoder_class('Signature', metadata=self.metadata)
            data += self.signature.encode('0x{}'.format(value['signature'].replace('0x', '')))

            self.era = self.get_decoder_class('Era', metadata=self.metadata)
            data += self.era.encode(value['era'])

            self.nonce = self.get_decoder_class('Compact<Index>', metadata=self.metadata)
            data += self.nonce.encode(value['nonce'])

            self.tip = self.get_decoder_class('Compact<Balance>', metadata=self.metadata)
            data += self.tip.encode(value['tip'])

        else:
            data = ScaleBytes('0x04')

        data += ScaleBytes(bytearray.fromhex(self.call_index))

        # Convert params to call_args TODO refactor
        if not value.get('call_args') and value.get('params'):
            value['call_args'] = {call_arg['name']: call_arg['value'] for call_arg in value.get('params')}

        # Encode call params
        if len(self.call.args) > 0:
            for arg in self.call.args:
                if arg.name not in value.get('call_args', {}):
                    raise ValueError('Parameter \'{}\' not specified'.format(arg.name))
                else:
                    param_value = value['call_args'][arg.name]

                    arg_obj = self.get_decoder_class(arg.type, metadata=self.metadata)
                    data += arg_obj.encode(param_value)

        # Wrap payload with a length Compact<u32>
        length_obj = self.get_decoder_class('Compact<u32>')
        data = length_obj.encode(data.length) + data

        return data


class ExtrinsicsBlock61181Decoder(ExtrinsicsDecoder):
    type_mapping = (
        ('extrinsic_length', 'Compact<u32>'),
        ('version_info', 'u8'),
        ('address', 'Address'),
        ('signature', 'Signature'),
        ('nonce', 'u64'),
        ('era', 'Era'),
        ('call_index', '(u8,u8)'),
    )


class EventsDecoder(Vec):
    type_string = 'Vec<EventRecord<Event, Hash>>'

    def __init__(self, data, metadata=None, **kwargs):
        assert (not metadata or type(metadata) == MetadataDecoder)

        self.metadata = metadata
        self.elements = []

        super().__init__(data, metadata=metadata, **kwargs)

    def process(self):
        element_count = self.process_type('Compact<u32>').value

        for i in range(0, element_count):
            element = self.process_type('EventRecord', metadata=self.metadata)
            element.value['event_idx'] = i
            self.elements.append(element)

        return [e.value for e in self.elements]


class EventRecord(ScaleDecoder):

    def __init__(self, data, sub_type=None, metadata: MetadataDecoder = None):

        assert (not metadata or type(metadata) == MetadataDecoder)

        self.metadata = metadata

        self.phase = None
        self.extrinsic_idx = None
        self.type = None
        self.params = []
        self.event = None
        self.event_module = None
        self.topics = []

        super().__init__(data, sub_type)

    def process(self):

        # TODO Create option type
        self.phase = self.get_next_u8()

        if self.phase == 0:
            self.extrinsic_idx = self.process_type('U32').value

        self.type = self.get_next_bytes(2).hex()

        # Decode params

        self.event = self.metadata.event_index[self.type][1]
        self.event_module = self.metadata.event_index[self.type][0]

        for arg_type in self.event.args:
            arg_type_obj = self.process_type(arg_type)

            self.params.append({
                'type': arg_type,
                'value': arg_type_obj.serialize(),
                'valueRaw': arg_type_obj.raw_value
            })

        # Topics introduced since MetadataV5
        if self.metadata.version and self.metadata.version.index >= 5:
            self.topics = self.process_type('Vec<Hash>').value

        return {
            'phase': self.phase,
            'extrinsic_idx': self.extrinsic_idx,
            'type': self.type,
            'module_id': self.event_module.name,
            'event_id': self.event.name,
            'params': self.params,
            'topics': self.topics
        }


class Other(Bytes):
    pass


class AuthoritiesChange(Vec):
    type_string = 'Vec<AccountId>'

    def __init__(self, data, **kwargs):

        super().__init__(data, sub_type='AccountId', **kwargs)


class ChangesTrieRoot(Bytes):
    pass


class SealV0(Struct):
    type_string = '(u64, Signature)'

    type_mapping = (('slot', 'u64'), ('signature', 'Signature'))


class Consensus(Struct):
    type_string = '(ConsensusEngineId, Vec<u8>)'

    type_mapping = (('engine', 'ConsensusEngineId'), ('data', 'HexBytes'))


class Seal(Struct):
    type_string = '(ConsensusEngineId, Bytes)'

    type_mapping = (('engine', 'ConsensusEngineId'), ('data', 'HexBytes'))


class PreRuntime(Struct):
    type_string = '(ConsensusEngineId, Bytes)'

    type_mapping = (('engine', 'ConsensusEngineId'), ('data', 'HexBytes'))


class LogDigest(Enum):

    value_list = ['Other', 'AuthoritiesChange', 'ChangesTrieRoot', 'SealV0', 'Consensus', 'Seal', 'PreRuntime']

    def __init__(self, data, **kwargs):
        self.log_type = None
        self.index_value = None
        super().__init__(data, **kwargs)

    def process(self):
        self.index = int(self.get_next_bytes(1).hex())
        self.index_value = self.value_list[self.index]
        self.log_type = self.process_type(self.value_list[self.index])

        return {'type': self.log_type.type_string, 'value': self.log_type.value}




#  Polkascan Substrate Interface GUI
#
#  Copyright 2018-2020 openAware BV (NL).
#  This file is part of Polkascan.
#
#  Polkascan is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Polkascan is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Polkascan. If not, see <http://www.gnu.org/licenses/>.
#
#  test_scalebytes.py
#

import unittest

from scalecodec.base import ScaleDecoder, ScaleBytes
from scalecodec.exceptions import RemainingScaleBytesNotEmptyException


class TestScaleBytes(unittest.TestCase):

    def test_unknown_data_format(self):
        self.assertRaises(ValueError, ScaleBytes, 123)
        self.assertRaises(ValueError, ScaleBytes, 'test')

    def test_bytes_data_format(self):
        obj = ScaleDecoder.get_decoder_class('Compact<u32>', ScaleBytes(b"\x02\x09\x3d\x00"))
        obj.decode()
        self.assertEqual(obj.value, 1000000)

    def test_remaining_bytes(self):
        scale = ScaleBytes("0x01020304")
        scale.get_next_bytes(1)
        self.assertEqual(scale.get_remaining_bytes(), b'\x02\x03\x04')

    def test_reset(self):
        scale = ScaleBytes("0x01020304")
        scale.get_next_bytes(1)
        scale.reset()
        self.assertEqual(scale.get_remaining_bytes(), b'\x01\x02\x03\x04')

    def test_abstract_process(self):
        self.assertRaises(NotImplementedError, ScaleDecoder.process, None)

    def test_abstract_encode(self):
        self.assertRaises(NotImplementedError, ScaleDecoder.process_encode, None, None)

    def test_add_scalebytes(self):
        scale_total = ScaleBytes("0x0102") + "0x0304"

        self.assertEqual(scale_total.data, bytearray.fromhex("01020304"))

    def test_scale_decoder_remaining_bytes(self):
        obj = ScaleDecoder.get_decoder_class('[u8; 3]', ScaleBytes("0x010203"))
        self.assertEqual(obj.get_remaining_bytes(), b"\x01\x02\x03")

    def test_no_more_bytes_available(self):
        obj = ScaleDecoder.get_decoder_class('[u8; 4]', ScaleBytes("0x010203"))
        self.assertRaises(RemainingScaleBytesNotEmptyException, obj.decode, False)

    def test_str_representation(self):
        obj = ScaleDecoder.get_decoder_class('Bytes', ScaleBytes("0x1054657374"))
        obj.decode()
        self.assertEqual(str(obj), "Test")

    def test_type_convert(self):
        self.assertEqual(ScaleDecoder.convert_type("RawAddress"), "Address")
        self.assertEqual(ScaleDecoder.convert_type("<Balance as HasCompact>::Type"), "Compact<Balance>")
        self.assertEqual(ScaleDecoder.convert_type("<BlockNumber as HasCompact>::Type"), "Compact<BlockNumber>")
        self.assertEqual(ScaleDecoder.convert_type("<Moment as HasCompact>::Type"), "Compact<Moment>")




# Python SCALE Codec Library
#
# Copyright 2018-2020 openAware BV (NL).
# This file is part of Polkascan.
#
# Polkascan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polkascan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polkascan. If not, see <http://www.gnu.org/licenses/>.

import unittest

from scalecodec.base import ScaleBytes, ScaleDecoder, RuntimeConfiguration
from scalecodec.metadata import MetadataDecoder
from scalecodec.type_registry import load_type_registry_preset

from scalecodec.types import CompactU32, Vec
from test.fixtures import metadata_v10_hex


class TestScaleTypeEncoding(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        RuntimeConfiguration().update_type_registry(load_type_registry_preset("default"))
        RuntimeConfiguration().update_type_registry(load_type_registry_preset("kusama"))

        cls.metadata_decoder = MetadataDecoder(ScaleBytes(metadata_v10_hex))
        cls.metadata_decoder.decode()

    def test_u16(self):
        obj = ScaleDecoder.get_decoder_class('u16')
        obj.encode(64302)
        self.assertEqual(str(obj.data), "0x2efb")

    def test_i16(self):
        obj = ScaleDecoder.get_decoder_class('i16')
        obj.encode(-1234)
        self.assertEqual(str(obj.data), "0x2efb")

    def test_i16_out_of_bounds(self):
        obj = ScaleDecoder.get_decoder_class('i16')
        self.assertRaises(ValueError, obj.encode, -32769)

    def test_compact_u32_1byte(self):
        obj = ScaleDecoder.get_decoder_class('Compact<u32>', ScaleBytes("0x18"))
        obj.decode()

        obj = ScaleDecoder.get_decoder_class('Compact<u32>')
        obj.encode(6)
        self.assertEqual(str(obj.data), "0x18")

    def test_compact_u32_2bytes(self):
        obj = ScaleDecoder.get_decoder_class('Compact<u32>', ScaleBytes("0x18"))
        obj.decode()

        obj = ScaleDecoder.get_decoder_class('Compact<u32>')
        obj.encode(6000)
        self.assertEqual(str(obj.data), "0xc15d")

    def test_compact_u32_4bytes(self):

        obj = ScaleDecoder.get_decoder_class('Compact<u32>')
        obj.encode(1000000)
        self.assertEqual(str(obj.data), "0x02093d00")

    def test_compact_u32_larger_than_4bytes(self):

        obj = ScaleDecoder.get_decoder_class('Compact<u32>')
        obj.encode(150000000000000)
        self.assertEqual(str(obj.data), "0x0b0060b7986c88")

    def test_compact_u32_encode_decode(self):

        value = 2000001

        obj = ScaleDecoder.get_decoder_class('Compact<u32>')
        data = obj.encode(value)

        obj = CompactU32(data)

        self.assertEqual(obj.decode(), value)

    def test_compact_u32_encode_decode_large(self):

        value = 2**30

        obj = CompactU32(ScaleBytes(bytearray()))
        data = obj.encode(value)

        obj = CompactU32(data)

        self.assertEqual(obj.decode(), value)

    def test_vec_string_encode_decode(self):

        value = ['test', 'vec']

        obj = ScaleDecoder.get_decoder_class('Vec<Bytes>')
        data = obj.encode(value)

        obj = ScaleDecoder.get_decoder_class('Vec<Bytes>', data)

        self.assertEqual(obj.decode(), value)

    def test_vec_accountid_encode_decode(self):

        value = [
            '0x0034d9d2dcdcd79451d95fd019a056d47dfa9926d762b94e63f06391b1545aee',
            '0x2ce1929ab903f695bdeeeb79a588774d71468362129136f1b7f7b31a32958f98',
            '0x88c47944e4aaf9d53a9627400f9a948bb5f355bda38702dbdeda0c5d34553128',
        ]

        obj = ScaleDecoder.get_decoder_class('Vec<AccountId>')
        data = obj.encode(value)

        obj = ScaleDecoder.get_decoder_class('Vec<AccountId>', data)

        self.assertEqual(obj.decode(), value)

    def test_bytes_encode_decode(self):

        value = 'This is a test'

        obj = ScaleDecoder.get_decoder_class('Bytes')
        data = obj.encode(value)

        obj_check = ScaleDecoder.get_decoder_class('Bytes', data)

        self.assertEqual(obj_check.decode(), value)

    def test_hexbytes_encode_decode(self):

        value = '0x5468697320697320612074657374'

        obj = ScaleDecoder.get_decoder_class('HexBytes')
        data = obj.encode(value)

        obj_check = ScaleDecoder.get_decoder_class('HexBytes', data)

        self.assertEqual(obj_check.decode(), value)

    def test_accountid_encode_decode(self):
        value = '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'

        obj = ScaleDecoder.get_decoder_class('AccountId')
        data = obj.encode(value)

        obj_check = ScaleDecoder.get_decoder_class('AccountId', data)

        self.assertEqual(obj_check.decode(), value)

    def test_compact_balance_encode_decode(self):
        scale_data = ScaleBytes('0x070010a5d4e8')
        value = 1000000000000

        obj = ScaleDecoder.get_decoder_class('Compact<Balance>')
        data = obj.encode(value)

        self.assertEqual(str(scale_data), str(data))

        obj_check = ScaleDecoder.get_decoder_class('Compact<Balance>', data)

        self.assertEqual(obj_check.decode(), value)

    def test_struct_encode_decode(self):

        value = {'unstakeThreshold': 3, 'validatorPayment': 0}
        scale_data = ScaleBytes("0x0c00")

        obj = ScaleDecoder.get_decoder_class('ValidatorPrefsLegacy')
        data = obj.encode(value)

        self.assertEqual(str(scale_data), str(data))

        obj_check = ScaleDecoder.get_decoder_class('ValidatorPrefsLegacy', data)

        self.assertEqual(obj_check.decode(), value)

    def test_enum_encode_decode(self):

        value = "Stash"

        obj = ScaleDecoder.get_decoder_class('RewardDestination')
        data = obj.encode(value)

        obj_check = ScaleDecoder.get_decoder_class('RewardDestination', data)

        self.assertEqual(obj_check.decode(), value)

    def test_enum_type_mapping_encode_decode(self):
        RuntimeConfiguration().update_type_registry(load_type_registry_preset("test"))

        value = {"AuthoritiesChange": ["0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409"]}

        obj = ScaleDecoder.get_decoder_class('DigestItem')
        data = obj.encode(value)

        obj_check = ScaleDecoder.get_decoder_class('DigestItem', data)

        self.assertEqual(obj_check.decode(), value)

    def test_option_empty_encode_decode(self):

        value = None

        obj = ScaleDecoder.get_decoder_class('Option<Bytes>')
        data = obj.encode(value)

        obj_check = ScaleDecoder.get_decoder_class('Option<Bytes>', data)

        self.assertEqual(obj_check.decode(), value)

    def test_option_bytes_encode_decode(self):
        value = "Test"

        obj = ScaleDecoder.get_decoder_class('Option<Bytes>')
        data = obj.encode(value)

        obj_check = ScaleDecoder.get_decoder_class('Option<Bytes>', data)

        self.assertEqual(obj_check.decode(), value)

    def test_proposal_encode_decode(self):

        value = {
            'call_module': 'System',
            'call_function': 'remark',
            'call_args': {
                '_remark': '0x0123456789'
            }
        }

        obj = ScaleDecoder.get_decoder_class('Box<Proposal>', metadata=self.metadata_decoder)
        data = obj.encode(value)

        obj_check = ScaleDecoder.get_decoder_class('Box<Proposal>', data, metadata=self.metadata_decoder)

        obj_check.decode()

        self.assertEqual(obj_check.value['call_module'], 'System')
        self.assertEqual(obj_check.value['call_function'], 'remark')
        self.assertEqual(obj_check.value['call_args'][0]['value'], '0x0123456789')

    def test_set_encode_decode(self):

        RuntimeConfiguration().update_type_registry(load_type_registry_preset("test"))

        value = ['Display', 'Legal', 'Email', 'Twitter']

        obj = ScaleDecoder.get_decoder_class('IdentityFields')
        scale_data = obj.encode(value)

        obj = ScaleDecoder.get_decoder_class('IdentityFields', scale_data)
        obj.decode()

        self.assertEqual(obj.value, value)

    def test_data_encode_decode(self):

        value = {"Raw": "Test"}

        obj = ScaleDecoder.get_decoder_class('Data')
        scale_data = obj.encode(value)

        obj = ScaleDecoder.get_decoder_class('Data', scale_data)
        obj.decode()

        self.assertEqual(obj.value, value)





# Python SCALE Codec Library
#
# Copyright 2018-2020 openAware BV (NL).
# This file is part of Polkascan.
#
# Polkascan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Polkascan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Polkascan. If not, see <http://www.gnu.org/licenses/>.

import unittest

from scalecodec.base import ScaleBytes, ScaleDecoder, RuntimeConfiguration
from scalecodec.block import ExtrinsicsDecoder
from scalecodec.metadata import MetadataDecoder
from scalecodec.type_registry import load_type_registry_preset

from scalecodec.types import CompactU32, Vec
from test.fixtures import metadata_1045_hex


class TestScaleTypeEncoding(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        RuntimeConfiguration().update_type_registry(load_type_registry_preset("default"))
        RuntimeConfiguration().update_type_registry(load_type_registry_preset("kusama"))
        RuntimeConfiguration().set_active_spec_version_id(1045)

        cls.metadata_decoder = MetadataDecoder(ScaleBytes(metadata_1045_hex))
        cls.metadata_decoder.decode()

    def test_decode_balance_transfer_payload(self):
        unsigned_payload = "0xa8040400ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e8"

        extrinsic = ExtrinsicsDecoder(
            data=ScaleBytes(unsigned_payload),
            metadata=self.metadata_decoder
        )
        extrinsic.decode()

        call_module, call_function = self.metadata_decoder.call_index[extrinsic.call_index]

        # Check call module
        self.assertEqual(call_module.name, 'Balances')

        # Check call function
        self.assertEqual(call_function.name, 'transfer')

        # Check destination address for balance transfer
        self.assertEqual(extrinsic.params[0]['type'], 'Address')
        self.assertEqual(extrinsic.params[0]['value'],
                         '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409')

        # Check value of balance transfer
        self.assertEqual(extrinsic.params[1]['type'], 'Compact<Balance>')
        self.assertEqual(extrinsic.params[1]['value'], 1000000000000)

    def test_encode_attestations_more_attestations_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Attestations',
            'call_function': 'more_attestations',
            'call_args': {
                '_more': {}
            }
        })

        self.assertEqual(str(payload), "0x0c041500")

    # TODO: complicated: {"parent_hash":"0x0000000000000000000000000000000000000000000000000000000000000000","number":0,"state_root":"0x0000000000000000000000000000000000000000000000000000000000000000","extrinsics_root":"0x0000000000000000000000000000000000000000000000000000000000000000","digest":{"logs":["0x0000"]}}
    def test_encode_authorship_set_uncles_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Authorship',
            'call_function': 'set_uncles',
            'call_args': {
                'new_uncles': [
                    {
                        "parent_hash": "0x0000000000000000000000000000000000000000000000000000000000000000",
                        "number": 0,
                        "state_root": "0x0000000000000000000000000000000000000000000000000000000000000000",
                        "extrinsics_root": "0x0000000000000000000000000000000000000000000000000000000000000000",
                        "digest": {"logs": []}
                    }
                ]
            }
        })

        self.assertEqual(str(payload),
                         "0x9901040500040000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000")

    def test_encode_balances_force_transfer_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Balances',
            'call_function': 'force_transfer',
            'call_args': {
                'dest': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409',
                'source': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409',
                'value': 1000000000000
            }
        })

        self.assertEqual(str(payload),
                         "0x2d01040402ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e8")

    def test_encode_balances_force_transfer_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Balances',
            'call_function': 'force_transfer',
            'call_args': {
                'dest': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk',
                'source': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk',
                'value': 1000000000000
            }
        })

        self.assertEqual(str(payload),
                         "0x2d01040402ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e8")

    def test_encode_balances_set_balance_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Balances',
            'call_function': 'set_balance',
            'call_args': {
                'new_free': 1000000000000,
                'new_reserved': 2000000000000,
                'who': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload),
                         "0xc4040401ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e80b00204aa9d101")

    def test_encode_balances_set_balance_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Balances',
            'call_function': 'set_balance',
            'call_args': {
                'new_free': 1000000000000,
                'new_reserved': 2000000000000,
                'who': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload),
                         "0xc4040401ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e80b00204aa9d101")

    def test_encode_balances_transfer_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Balances',
            'call_function': 'transfer',
            'call_args': {
                'dest': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409',
                'value': 1000000000000
            }
        })

        self.assertEqual(str(payload),
                         "0xa8040400ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e8")

    def test_encode_balance_transfer_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Balances',
            'call_function': 'transfer',
            'call_args': {
                'dest': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk',
                'value': 1000000000000
            }
        })

        self.assertEqual(str(payload),
                         "0xa8040400ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e8")

    def test_encode_balances_transfer_keep_alive_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Balances',
            'call_function': 'transfer_keep_alive',
            'call_args': {
                'dest': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409',
                'value': 1000000000000
            }
        })

        self.assertEqual(str(payload),
                         "0xa8040403ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e8")

    def test_encode_balances_transfer_keep_alive_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Balances',
            'call_function': 'transfer_keep_alive',
            'call_args': {
                'dest': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk',
                'value': 1000000000000
            }
        })

        self.assertEqual(str(payload),
                         "0xa8040403ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e8")

    def test_encode_claims_claim_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Claims',
            'call_function': 'claim',
            'call_args': {
                'dest': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409',
                'ethereum_signature': '0xd7c4955996cf00953e65ec1895825b9c3894041ed8ab6bd671c456d53f5d04c13948a58a5f20c7c0d3f1e0d08c33ff590a8c681f6a9db78477ca83c8ab8711f500'
            }
        })

        self.assertEqual(str(payload),
                         "0x9101041300586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409d7c4955996cf00953e65ec1895825b9c3894041ed8ab6bd671c456d53f5d04c13948a58a5f20c7c0d3f1e0d08c33ff590a8c681f6a9db78477ca83c8ab8711f500")

    def test_encode_claims_claim_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Claims',
            'call_function': 'claim',
            'call_args': {
                'dest': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk',
                'ethereum_signature': '0xd7c4955996cf00953e65ec1895825b9c3894041ed8ab6bd671c456d53f5d04c13948a58a5f20c7c0d3f1e0d08c33ff590a8c681f6a9db78477ca83c8ab8711f500'
            }
        })

        self.assertEqual(str(payload),
                         "0x9101041300586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409d7c4955996cf00953e65ec1895825b9c3894041ed8ab6bd671c456d53f5d04c13948a58a5f20c7c0d3f1e0d08c33ff590a8c681f6a9db78477ca83c8ab8711f500")

    def test_encode_claims_mint_claim_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Claims',
            'call_function': 'mint_claim',
            'call_args': {
                'value': 1000000000000,
                'vesting_schedule': None,
                'who': '0x0123456789012345678901234567890123456789'
            }
        })

        self.assertEqual(str(payload),
                         "0xa004130101234567890123456789012345678901234567890010a5d4e8000000000000000000000000")

    def test_encode_claims_mint_claim_withvesting_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Claims',
            'call_function': 'mint_claim',
            'call_args': {
                'value': 1000000000000,
                'vesting_schedule': [2000000000000, 3000000000000, 1000000],
                'who': '0x0123456789012345678901234567890123456789'
            }
        })

        self.assertEqual(str(payload),
                         "0x310104130101234567890123456789012345678901234567890010a5d4e800000000000000000000000100204aa9d101000000000000000000000030ef7dba020000000000000000000040420f00")

    def test_encode_council_execute_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Council',
            'call_function': 'execute',
            'call_args': {
                'proposal': {
                    'call_module': 'System',
                    'call_function': 'remark',
                    'call_args': {
                        '_remark': '0x0123456789'
                    }
                }
            }
        })

        self.assertEqual(str(payload), "0x2c040e010001140123456789")

    def test_encode_council_propose_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Council',
            'call_function': 'propose',
            'call_args': {
                'proposal': {
                    'call_module': 'System',
                    'call_function': 'remark',
                    'call_args': {
                        '_remark': '0x0123456789'
                    }
                },
                'threshold': 7
            }
        })

        self.assertEqual(str(payload), "0x30040e021c0001140123456789")

    def test_encode_council_set_members_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Council',
            'call_function': 'set_members',
            'call_args': {
                'new_members': ["0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409",
                                "0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409"]
            }
        })

        self.assertEqual(str(payload),
                         "0x1101040e0008586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_council_set_members_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Council',
            'call_function': 'set_members',
            'call_args': {
                'new_members': ["EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk",
                                "EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk"]
            }
        })

        self.assertEqual(str(payload),
                         "0x1101040e0008586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_council_vote_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Council',
            'call_function': 'vote',
            'call_args': {
                'approve': True,
                'index': 1,
                'proposal': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
            }
        })

        self.assertEqual(str(payload), "0x94040e030123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0401")

    def test_encode_democracy_cancel_queued_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'cancel_queued',
            'call_args': {
                'which': 1
            }
        })

        self.assertEqual(str(payload), "0x1c040d0b01000000")

    def test_encode_democracy_cancel_referendum_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'cancel_referendum',
            'call_args': {
                'ref_index': 1
            }
        })

        self.assertEqual(str(payload), "0x10040d0a04")

    def test_encode_democracy_clear_public_proposals_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'clear_public_proposals',
            'call_args': {
            }
        })

        self.assertEqual(str(payload), "0x0c040d11")

    def test_encode_democracy_delegate_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'delegate',
            'call_args': {
                'conviction': 'None',
                'to': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x90040d0f586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c40900")

    def test_encode_democracy_delegate_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'delegate',
            'call_args': {
                'conviction': 'None',
                'to': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x90040d0f586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c40900")

    def test_encode_democracy_delegate_withconviction_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'delegate',
            'call_args': {
                'conviction': 'Locked5x',
                'to': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x90040d0f586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c40905")

    def test_encode_democracy_delegate_withconviction_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'delegate',
            'call_args': {
                'conviction': 'Locked5x',
                'to': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x90040d0f586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c40905")

    def test_encode_democracy_emergency_cancel_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'emergency_cancel',
            'call_args': {
                'ref_index': 1
            }
        })

        self.assertEqual(str(payload), "0x1c040d0401000000")

    def test_encode_democracy_external_propose_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'external_propose',
            'call_args': {
                'proposal_hash': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
            }
        })

        self.assertEqual(str(payload), "0x8c040d050123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef")

    def test_encode_democracy_external_propose_default_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'external_propose_default',
            'call_args': {
                'proposal_hash': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
            }
        })

        self.assertEqual(str(payload), "0x8c040d070123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef")

    def test_encode_democracy_external_propose_majority_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'external_propose_majority',
            'call_args': {
                'proposal_hash': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
            }
        })

        self.assertEqual(str(payload), "0x8c040d060123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef")

    def test_encode_democracy_fast_track_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'fast_track',
            'call_args': {
                'delay': 2000,
                'proposal_hash': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
                'voting_period': 1000
            }
        })

        self.assertEqual(str(payload),
                         "0xac040d080123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdefe8030000d0070000")

    def test_encode_democracy_note_imminent_preimage_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'note_imminent_preimage',
            'call_args': {
                'encoded_proposal': '0x00'
            }
        })

        self.assertEqual(str(payload), "0x14040d130400")

    def test_encode_democracy_note_preimage_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'note_preimage',
            'call_args': {
                'encoded_proposal': '0x00'
            }
        })

        self.assertEqual(str(payload), "0x14040d120400")

    def test_encode_democracy_propose_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'propose',
            'call_args': {
                'proposal_hash': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
                'value': 1000000000000
            }
        })

        self.assertEqual(str(payload),
                         "0xa4040d000123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef070010a5d4e8")

    def test_encode_democracy_proxy_vote_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'proxy_vote',
            'call_args': {
                'ref_index': 0,
                'vote': 128
            }
        })

        self.assertEqual(str(payload), "0x14040d030080")

    def test_encode_democracy_reap_preimage_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'reap_preimage',
            'call_args': {
                'proposal_hash': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
            }
        })

        self.assertEqual(str(payload), "0x8c040d140123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef")

    def test_encode_democracy_remove_proxy_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'remove_proxy',
            'call_args': {
                'proxy': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x8c040d0e586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_democracy_remove_proxy_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'remove_proxy',
            'call_args': {
                'proxy': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x8c040d0e586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_democracy_resign_proxy_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'resign_proxy',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c040d0d")

    def test_encode_democracy_second_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'second',
            'call_args': {
                'proposal': 1
            }
        })

        self.assertEqual(str(payload), "0x10040d0104")

    def test_encode_democracy_set_proxy_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'set_proxy',
            'call_args': {
                'proxy': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x8c040d0c586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_democracy_set_proxy_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'set_proxy',
            'call_args': {
                'proxy': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x8c040d0c586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_democracy_undelegate_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'undelegate',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c040d10")

    def test_encode_democracy_veto_external_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'veto_external',
            'call_args': {
                'proposal_hash': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
            }
        })

        self.assertEqual(str(payload), "0x8c040d090123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef")

    def test_encode_democracy_vote_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Democracy',
            'call_function': 'vote',
            'call_args': {
                'ref_index': 0,
                'vote': 128
            }
        })

        self.assertEqual(str(payload), "0x14040d020080")

    def test_encode_electionsphragmen_remove_member_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'ElectionsPhragmen',
            'call_function': 'remove_member',
            'call_args': {
                'who': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x90041005ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_electionsphragmen_remove_member_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'ElectionsPhragmen',
            'call_function': 'remove_member',
            'call_args': {
                'who': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x90041005ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_electionsphragmen_remove_voter_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'ElectionsPhragmen',
            'call_function': 'remove_voter',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c041001")

    def test_encode_electionsphragmen_renounce_candidacy_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'ElectionsPhragmen',
            'call_function': 'renounce_candidacy',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c041004")

    def test_encode_electionsphragmen_report_defunct_voter_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'ElectionsPhragmen',
            'call_function': 'report_defunct_voter',
            'call_args': {
                'target': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x90041002ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_electionsphragmen_report_defunct_voter_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'ElectionsPhragmen',
            'call_function': 'report_defunct_voter',
            'call_args': {
                'target': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x90041002ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_electionsphragmen_submit_candidacy_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'ElectionsPhragmen',
            'call_function': 'submit_candidacy',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c041003")

    def test_encode_electionsphragmen_vote_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'ElectionsPhragmen',
            'call_function': 'vote',
            'call_args': {
                'value': 1000000000000,
                'votes': ["0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409",
                          "0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409",
                          "0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409"]
            }
        })

        self.assertEqual(str(payload),
                         "0xa9010410000c586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e8")

    def test_encode_electionsphragmen_vote_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'ElectionsPhragmen',
            'call_function': 'vote',
            'call_args': {
                'value': 1000000000000,
                'votes': ["EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk",
                          "EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk",
                          "EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk"]
            }
        })

        self.assertEqual(str(payload),
                         "0xa9010410000c586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e8")

    def test_encode_finalitytracker_final_hint_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'FinalityTracker',
            'call_function': 'final_hint',
            'call_args': {
                'hint': 500000
            }
        })

        self.assertEqual(str(payload), "0x1c04090082841e00")

    def test_encode_grandpa_report_misbehavior_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Grandpa',
            'call_function': 'report_misbehavior',
            'call_args': {
                '_report': '0x00'
            }
        })

        self.assertEqual(str(payload), "0x14040a000400")

    def test_encode_identity_add_registrar_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'add_registrar',
            'call_args': {
                'account': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x8c041900586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_identity_add_registrar_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'add_registrar',
            'call_args': {
                'account': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x8c041900586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_identity_cancel_request_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'cancel_request',
            'call_args': {
                'reg_index': 1
            }
        })

        self.assertEqual(str(payload), "0x1c04190501000000")

    def test_encode_identity_clear_identity_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'clear_identity',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c041903")

    def test_encode_identity_kill_identity_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'kill_identity',
            'call_args': {
                'target': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x9004190aff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_identity_kill_identity_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'kill_identity',
            'call_args': {
                'target': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x9004190aff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_identity_provide_judgement_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'provide_judgement',
            'call_args': {
                'judgement': {"KnownGood": None},
                'reg_index': 1,
                'target': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload),
                         "0x9804190904ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c40903")

    def test_encode_identity_provide_judgement_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'provide_judgement',
            'call_args': {
                'judgement': {"KnownGood": None},
                'reg_index': 1,
                'target': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload),
                         "0x9804190904ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c40903")

    def test_encode_identity_request_judgement_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'request_judgement',
            'call_args': {
                'max_fee': 2000000000000,
                'reg_index': 1
            }
        })

        self.assertEqual(str(payload), "0x2c041904040b00204aa9d101")

    def test_encode_identity_set_account_id_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'set_account_id',
            'call_args': {
                'index': 1,
                'new': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x9004190704586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_identity_set_account_id_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'set_account_id',
            'call_args': {
                'index': 1,
                'new': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x9004190704586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_identity_set_fee_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'set_fee',
            'call_args': {
                'fee': 1000000000000,
                'index': 1
            }
        })

        self.assertEqual(str(payload), "0x2804190604070010a5d4e8")

    # TODO: Unable to determine default type for {"info":6,"type":"IdentityFields"}
    def test_encode_identity_set_fields_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'set_fields',
            'call_args': {
                'fields': ['Display', 'Legal', 'Email', 'Twitter'],
                'index': 1
            }
        })

        self.assertEqual(str(payload), "0x30041908049300000000000000")

    def test_encode_identity_set_identity_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'set_identity',
            'call_args': {
                'info': {
                    "additional": [],
                    "display": {"Raw": "Test1"},
                    "legal": {"Raw": "Test2"},
                    "web": {"None": None},
                    "riot": {"Raw": "Test3"},
                    "email": {"None": None},
                    "pgpFingerprint": None,
                    "image": {"None": None},
                    "twitter": {"None": None}
                }
            }
        })

        self.assertEqual(str(payload),
                         "0x6c041901000654657374310654657374320006546573743300000000")

    def test_encode_identity_set_subs_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'set_subs',
            'call_args': {
                'subs': [
                    ["0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409", {"None": None}],
                    ["0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409", {"Raw": "Test"}]
                ]
            }
        })

        self.assertEqual(str(payload),
                         "0x290104190208586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c40900586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c4090554657374")

    def test_encode_identity_set_subs_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Identity',
            'call_function': 'set_subs',
            'call_args': {
                'subs': [
                    ["EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk", {"None": None}],
                    ["EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk", {"Raw": "Test"}]
                ]
            }
        })

        self.assertEqual(str(payload),
                         "0x290104190208586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c40900586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c4090554657374")

    def test_encode_imonline_heartbeat_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'ImOnline',
            'call_function': 'heartbeat',
            'call_args': {
                'heartbeat': {"blockNumber": 500000, "networkState": {"peerId": "0x012345", "externalAddresses": []},
                              "sessionIndex": 1, "authorityIndex": 3},
                '_signature': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
            }
        })

        self.assertEqual(str(payload),
                         "0x5101040b0020a107000c0123450001000000030000000123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef")

    def test_encode_parachains_set_heads_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Parachains',
            'call_function': 'set_heads',
            'call_args': {
                'heads': [{
                    'candidate': {
                        "parachainIndex": 1,
                        "collator": "EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk",
                        "signature": "0x1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
                        "headData": "0x012345",
                        "balanceUploads": [],
                        "egressQueueRoots": [],
                        "fees": 1000000000000,
                        "blockDataHash": "0x0000000000000000000000000000000000000000000000000000000000000000"
                    },
                    'validityVotes': []
                }]
            }
        })

        self.assertEqual(str(payload),
                         "0x5d020414000401000000586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c4091234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef0c01234500000010a5d4e8000000000000000000000000000000000000000000000000000000000000000000000000")

    def test_encode_registrar_deregister_para_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Registrar',
            'call_function': 'deregister_para',
            'call_args': {
                'id': 1
            }
        })

        self.assertEqual(str(payload), "0x1004170104")

    def test_encode_registrar_deregister_parathread_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Registrar',
            'call_function': 'deregister_parathread',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c041705")

    def test_encode_registrar_register_para_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Registrar',
            'call_function': 'register_para',
            'call_args': {
                'code': '0x00',
                'id': 1,
                'info': {"scheduling": "Always"},
                'initial_head_data': '0x01'
            }
        })

        self.assertEqual(str(payload), "0x24041700040004000401")

    def test_encode_registrar_register_parathread_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Registrar',
            'call_function': 'register_parathread',
            'call_args': {
                'code': '0x00',
                'initial_head_data': '0x01'
            }
        })

        self.assertEqual(str(payload), "0x1c04170304000401")

    def test_encode_registrar_select_parathread_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Registrar',
            'call_function': 'select_parathread',
            'call_args': {
                '_collator': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
                '_head_hash': '0x0000000000000000000000000000000000000000000000000000000000000000',
                '_id': 1
            }
        })

        self.assertEqual(str(payload),
                         "0x1101041704040123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0000000000000000000000000000000000000000000000000000000000000000")

    def test_encode_registrar_set_thread_count_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Registrar',
            'call_function': 'set_thread_count',
            'call_args': {
                'count': 1000
            }
        })

        self.assertEqual(str(payload), "0x1c041702e8030000")

    def test_encode_registrar_swap_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Registrar',
            'call_function': 'swap',
            'call_args': {
                'other': 1
            }
        })

        self.assertEqual(str(payload), "0x1004170604")

    def test_encode_session_set_keys_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Session',
            'call_function': 'set_keys',
            'call_args': {
                'keys': {
                    "grandpa": "EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk",
                    "babe": "EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk",
                    "im_online": "EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk",
                    "authority_discovery": "EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk",
                    "parachains": "EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk"
                },
                'proof': '0x01234567890abcdef01234567890abcdef01234567890abcdef01234567890abcdef01234567890abcdef01234567890abcdef01234567890abcdef01234567890abcdef'
            }
        })

        self.assertEqual(str(payload),
                         "0xa503040800586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409110101234567890abcdef01234567890abcdef01234567890abcdef01234567890abcdef01234567890abcdef01234567890abcdef01234567890abcdef01234567890abcdef")

    def test_encode_slots_bid_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Slots',
            'call_function': 'bid',
            'call_args': {
                'amount': 1000000000000000,
                'auction_index': 2,
                'first_slot': 3,
                'last_slot': 4,
                'sub': 1
            }
        })

        self.assertEqual(str(payload), "0x3c04160104080c100f0080c6a47e8d03")

    def test_encode_slots_bid_renew_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Slots',
            'call_function': 'bid_renew',
            'call_args': {
                'amount': 1000000000000000,
                'auction_index': 1,
                'first_slot': 2,
                'last_slot': 3
            }
        })

        self.assertEqual(str(payload), "0x3804160204080c0f0080c6a47e8d03")

    def test_encode_slots_elaborate_deploy_data_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Slots',
            'call_function': 'elaborate_deploy_data',
            'call_args': {
                'code': '0x00',
                'para_id': 1
            }
        })

        self.assertEqual(str(payload), "0x18041605040400")

    def test_encode_slots_fix_deploy_data_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Slots',
            'call_function': 'fix_deploy_data',
            'call_args': {
                'code_hash': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
                'initial_head_data': '0x00',
                'para_id': 2,
                'sub': 1
            }
        })

        self.assertEqual(str(payload),
                         "0x9c04160404080123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0400")

    def test_encode_slots_new_auction_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Slots',
            'call_function': 'new_auction',
            'call_args': {
                'duration': 100000,
                'lease_period_index': 2
            }
        })

        self.assertEqual(str(payload), "0x20041600821a060008")

    def test_encode_slots_set_offboarding_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Slots',
            'call_function': 'set_offboarding',
            'call_args': {
                'dest': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x90041603ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_slots_set_offboarding_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Slots',
            'call_function': 'set_offboarding',
            'call_args': {
                'dest': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x90041603ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_staking_bond_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'bond',
            'call_args': {
                'controller': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409',
                'payee': 'Staked',
                'value': 1000000000000
            }
        })

        self.assertEqual(str(payload),
                         "0xac040600ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e800")

    def test_encode_staking_bond_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'bond',
            'call_args': {
                'controller': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk',
                'payee': 'Staked',
                'value': 1000000000000
            }
        })

        self.assertEqual(str(payload),
                         "0xac040600ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e800")

    def test_encode_staking_bond_extra_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'bond_extra',
            'call_args': {
                'max_additional': 1000000000000
            }
        })

        self.assertEqual(str(payload), "0x24040601070010a5d4e8")

    def test_encode_staking_cancel_deferred_slash_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'cancel_deferred_slash',
            'call_args': {
                'era': 1,
                'slash_indices': [0, 1, 2]
            }
        })

        self.assertEqual(str(payload), "0x5004060f010000000c000000000100000002000000")

    def test_encode_staking_chill_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'chill',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c040606")

    def test_encode_staking_force_new_era_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'force_new_era',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c04060b")

    def test_encode_staking_force_new_era_always_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'force_new_era_always',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c04060e")

    def test_encode_staking_force_no_eras_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'force_no_eras',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c04060a")

    def test_encode_staking_force_unstake_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'force_unstake',
            'call_args': {
                'stash': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x8c04060d586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_staking_force_unstake_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'force_unstake',
            'call_args': {
                'stash': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x8c04060d586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_staking_nominate_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'nominate',
            'call_args': {
                'targets': ['0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409']
            }
        })

        self.assertEqual(str(payload), "0x9404060504ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_staking_nominate_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'nominate',
            'call_args': {
                'targets': ['EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk']
            }
        })

        self.assertEqual(str(payload), "0x9404060504ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_staking_set_controller_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'set_controller',
            'call_args': {
                'controller': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x90040608ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_staking_set_controller_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'set_controller',
            'call_args': {
                'controller': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x90040608ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_staking_set_invulnerables_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'set_invulnerables',
            'call_args': {
                'validators': ['0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409',
                               '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409']
            }
        })

        self.assertEqual(str(payload),
                         "0x110104060c08586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_staking_set_invulnerables_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'set_invulnerables',
            'call_args': {
                'validators': ['EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk',
                               'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk']
            }
        })

        self.assertEqual(str(payload),
                         "0x110104060c08586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_staking_set_payee_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'set_payee',
            'call_args': {
                'payee': 'Staked'
            }
        })

        self.assertEqual(str(payload), "0x1004060700")

    def test_encode_staking_set_validator_count_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'set_validator_count',
            'call_args': {
                'new': 150
            }
        })

        self.assertEqual(str(payload), "0x140406095902")

    def test_encode_staking_unbond_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'unbond',
            'call_args': {
                'value': 1000000000000
            }
        })

        self.assertEqual(str(payload), "0x24040602070010a5d4e8")

    def test_encode_staking_validate_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'validate',
            'call_args': {
                'prefs': {"commission": 100}
            }
        })

        self.assertEqual(str(payload), "0x140406049101")

    def test_encode_staking_withdraw_unbonded_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Staking',
            'call_function': 'withdraw_unbonded',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c040603")

    def test_encode_system_fill_block_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'System',
            'call_function': 'fill_block',
            'call_args': {

            }
        })

        self.assertEqual(str(payload), "0x0c040000")

    def test_encode_system_kill_prefix_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'System',
            'call_function': 'kill_prefix',
            'call_args': {
                'prefix': '0x012345'
            }
        })

        self.assertEqual(str(payload), "0x1c0400060c012345")

    def test_encode_system_kill_storage_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'System',
            'call_function': 'kill_storage',
            'call_args': {
                'keys': ['0x0123456789abcdef', '0x0123456789abcdef']
            }
        })

        self.assertEqual(str(payload), "0x5804000508200123456789abcdef200123456789abcdef")

    def test_encode_system_remark_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'System',
            'call_function': 'remark',
            'call_args': {
                '_remark': '0x012345'
            }
        })

        self.assertEqual(str(payload), "0x1c0400010c012345")

    def test_encode_system_set_code_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'System',
            'call_function': 'set_code',
            'call_args': {
                'new': '0x012345'
            }
        })

        self.assertEqual(str(payload), "0x1c0400030c012345")

    def test_encode_system_set_heap_pages_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'System',
            'call_function': 'set_heap_pages',
            'call_args': {
                'pages': 100
            }
        })

        self.assertEqual(str(payload), "0x2c0400026400000000000000")

    def test_encode_system_set_storage_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'System',
            'call_function': 'set_storage',
            'call_args': {
                'items': [['key', 'value']]
            }
        })

        self.assertEqual(str(payload), "0x38040004040c6b65791476616c7565")

    def test_encode_technicalcommittee_execute_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalCommittee',
            'call_function': 'execute',
            'call_args': {
                'proposal': {
                    'call_module': 'System',
                    'call_function': 'remark',
                    'call_args': {
                        '_remark': '0x0123456789'
                    }
                }
            }
        })

        self.assertEqual(str(payload), "0x2c040f010001140123456789")

    def test_encode_technicalcommittee_propose_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalCommittee',
            'call_function': 'propose',
            'call_args': {
                'proposal': {
                    'call_module': 'System',
                    'call_function': 'remark',
                    'call_args': {
                        '_remark': '0x0123456789'
                    }
                },
                'threshold': 7
            }
        })

        self.assertEqual(str(payload), "0x30040f021c0001140123456789")

    def test_encode_technicalcommittee_set_members_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalCommittee',
            'call_function': 'set_members',
            'call_args': {
                'new_members': ['0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409',
                                '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409']
            }
        })

        self.assertEqual(str(payload),
                         "0x1101040f0008586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_technicalcommittee_set_members_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalCommittee',
            'call_function': 'set_members',
            'call_args': {
                'new_members': ['EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk',
                                'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk']
            }
        })

        self.assertEqual(str(payload),
                         "0x1101040f0008586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_technicalcommittee_vote_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalCommittee',
            'call_function': 'vote',
            'call_args': {
                'approve': True,
                'index': 1,
                'proposal': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
            }
        })

        self.assertEqual(str(payload), "0x94040f030123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef0401")

    def test_encode_technicalmembership_add_member_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalMembership',
            'call_function': 'add_member',
            'call_args': {
                'who': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x8c041100586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_technicalmembership_add_member_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalMembership',
            'call_function': 'add_member',
            'call_args': {
                'who': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x8c041100586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_technicalmembership_change_key_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalMembership',
            'call_function': 'change_key',
            'call_args': {
                'new': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x8c041104586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_technicalmembership_change_key_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalMembership',
            'call_function': 'change_key',
            'call_args': {
                'new': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x8c041104586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_technicalmembership_remove_member_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalMembership',
            'call_function': 'remove_member',
            'call_args': {
                'who': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload), "0x8c041101586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_technicalmembership_remove_member_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalMembership',
            'call_function': 'remove_member',
            'call_args': {
                'who': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload), "0x8c041101586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_technicalmembership_reset_members_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalMembership',
            'call_function': 'reset_members',
            'call_args': {
                'members': ['0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409',
                            '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409']
            }
        })

        self.assertEqual(str(payload),
                         "0x110104110308586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_technicalmembership_reset_members_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalMembership',
            'call_function': 'reset_members',
            'call_args': {
                'members': ['EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk',
                            'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk']
            }
        })

        self.assertEqual(str(payload),
                         "0x110104110308586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_technicalmembership_swap_member_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalMembership',
            'call_function': 'swap_member',
            'call_args': {
                'add': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409',
                'remove': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409'
            }
        })

        self.assertEqual(str(payload),
                         "0x0d01041102586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_technicalmembership_swap_member_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'TechnicalMembership',
            'call_function': 'swap_member',
            'call_args': {
                'add': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk',
                'remove': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'
            }
        })

        self.assertEqual(str(payload),
                         "0x0d01041102586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_timestamp_set_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Timestamp',
            'call_function': 'set',
            'call_args': {
                'now': 1
            }
        })

        self.assertEqual(str(payload), "0x1004020004")

    def test_encode_treasury_approve_proposal_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Treasury',
            'call_function': 'approve_proposal',
            'call_args': {
                'proposal_id': 1
            }
        })

        self.assertEqual(str(payload), "0x1004120204")

    def test_encode_treasury_propose_spend_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Treasury',
            'call_function': 'propose_spend',
            'call_args': {
                'beneficiary': '0x586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409',
                'value': 1000000000000
            }
        })

        self.assertEqual(str(payload),
                         "0xa8041200070010a5d4e8ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_treasury_propose_spend_ss58_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Treasury',
            'call_function': 'propose_spend',
            'call_args': {
                'beneficiary': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk',
                'value': 1000000000000
            }
        })

        self.assertEqual(str(payload),
                         "0xa8041200070010a5d4e8ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409")

    def test_encode_treasury_reject_proposal_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Treasury',
            'call_function': 'reject_proposal',
            'call_args': {
                'proposal_id': 1
            }
        })

        self.assertEqual(str(payload), "0x1004120104")

    def test_encode_utility_approve_as_multi_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Utility',
            'call_function': 'approve_as_multi',
            'call_args': {
                'call_hash': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
                'maybe_timepoint': {
                    'height': 444,
                    'index': 10
                },
                'other_signatories': ['EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk'],
                'threshold': 4
            }
        })

        self.assertEqual(str(payload), "0x3d01041803040004586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c40901bc0100000a0000000123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef")

    def test_encode_utility_as_multi_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Utility',
            'call_function': 'as_multi',
            'call_args': {
                'call': {
                    'call_module': 'System',
                    'call_function': 'remark',
                    'call_args': {
                        '_remark': '0x0123456789'
                    }
                },
                'maybe_timepoint': None,
                'other_signatories': [],
                'threshold': 5
            }
        })

        self.assertEqual(str(payload),
                         "0x3c041802050000000001140123456789")

    def test_encode_utility_as_sub_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Utility',
            'call_function': 'as_sub',
            'call_args': {
                'call': {
                    'call_module': 'System',
                    'call_function': 'remark',
                    'call_args': {
                        '_remark': '0x0123456789'
                    }
                },
                'index': 2
            }
        })

        self.assertEqual(str(payload), "0x3404180102000001140123456789")

    def test_encode_utility_batch_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Utility',
            'call_function': 'batch',
            'call_args': {
                'calls': [{
                    'call_module': 'Balances',
                    'call_function': 'transfer',
                    'call_args': {
                        'dest': 'EaG2CRhJWPb7qmdcJvy3LiWdh26Jreu9Dx6R1rXxPmYXoDk',
                        'value': 1000000000000
                    }
                }]
            }
        })

        self.assertEqual(str(payload), "0xb4041800040400ff586cb27c291c813ce74e86a60dad270609abf2fc8bee107e44a80ac00225c409070010a5d4e8")

    def test_encode_utility_cancel_as_multi_payload(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        payload = extrinsic.encode({
            'call_module': 'Utility',
            'call_function': 'cancel_as_multi',
            'call_args': {
                'call_hash': '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef',
                'other_signatories': [],
                'threshold': 5,
                'timepoint': {
                    'height': 10000,
                    'index': 1
                }
            }
        })

        self.assertEqual(str(payload), "0xb804180405000010270000010000000123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef")

    def test_signed_extrinsic(self):
        extrinsic = ExtrinsicsDecoder(metadata=self.metadata_decoder)

        extrinsic_value = {
            'account_id': '5E9oDs9PjpsBbxXxRE9uMaZZhnBAV38n2ouLB28oecBDdeQo',
            'signature_version': 1,
            'signature': '728b4057661816aa24918219ff90d10a34f1db4e81494d23c83ef54991980f77cf901acd970cb36d3c9c9e166d27a83a3aee648d4085e2bdb9e7622c0538e381',
            'call_function': 'transfer',
            'call_module': 'balances',
            'nonce': 0,
            'era': '00',
            'tip': 0,
            'params': [
                {'name': 'dest', 'type': 'Address',
                 'value': '0xd43593c715fdd31c61141abd04a99fd6822c8558854ccde39a5684e7a56da27d'},
                {'name': 'value', 'type': 'Compact<Balance>', 'value': 1000000000000}
            ]
        }

        extrinsic_hex = extrinsic.encode(extrinsic_value)

        obj = ScaleDecoder.get_decoder_class(
            "ExtrinsicsDecoder",
            data=extrinsic_hex,
            metadata=self.metadata_decoder
        )

        decoded_extrinsic = obj.decode()

        self.assertEqual(extrinsic_value['signature'], decoded_extrinsic['signature'])
        self.assertEqual(extrinsic_value['params'][0]['value'], decoded_extrinsic['params'][0]['value'])
